import logging
from typing import Optional, Union

from fastapi import FastAPI

from .instrumentation import IudexConfig
from .instrumentation import instrument as _instrument

logger = logging.getLogger(__name__)


def instrument_fastapi(
    app: FastAPI,
    service_name: Optional[str] = None,
    instance_id: Optional[str] = None,
    iudex_api_key: Optional[str] = None,
    log_level: Optional[Union[int, str]] = None,
    git_commit: Optional[str] = None,
    github_url: Optional[str] = None,
    env: Optional[str] = None,
    config: Optional[IudexConfig] = None,
):
    """Auto-instruments FastAPI app to send OTel signals to Iudex.

    Invoke this function in your FastAPI entrypoint.

    If you use Gunicorn, invoke it within a post_fork hook:
    ```python
    def post_fork(server, worker):
        instrument(...)
    ```
    Per https://opentelemetry-python.readthedocs.io/en/latest/examples/fork-process-model/README.html.

    Args:
        app: FastAPI app to instrument.
        service_name: Name of the service, e.g. "billing_service", __name__.
            If not supplied, env var OTEL_SERVICE_NAME will be used.
        instance_id: ID of the service instance, e.g. container id, pod name.
        iudex_api_key: Your Iudex API key.
            If not supplied, env var IUDEX_API_KEY will be used.
        log_level: Logging level for the root logger.
        git_commit: Commit hash of the currently deployed code.
            Used with github_url to deep link telemetry to source code.
        github_url: URL of the GitHub repository.
            Used with git_commit to deep link telemetry to source code.
        env: Environment of the service, e.g. "production", "staging".
        config: IudexConfig object with more granular options.
            Will override all other args, so provide them to the object instead.
    """
    iudex_config = _instrument(
        service_name=service_name,
        instance_id=instance_id,
        iudex_api_key=iudex_api_key,
        log_level=log_level,
        git_commit=git_commit,
        github_url=github_url,
        env=env,
        config=config,
    )

    try:
        from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

        instrumentor = FastAPIInstrumentor()
        if not instrumentor.is_instrumented_by_opentelemetry:
            instrumentor.instrument()
            instrumentor.instrument_app(app)
        return iudex_config
    except Exception as e:
        logger.error(f"Error initializing OpenAI instrumentor: {e}")

    return iudex_config



import logging
import os
from typing import Collection, Dict, Union

import postgrest._async.request_builder
import postgrest._sync.request_builder
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.semconv.attributes import http_attributes, server_attributes
from opentelemetry.semconv._incubating.attributes import db_attributes
from opentelemetry.trace import SpanKind, Status, StatusCode, Tracer, get_tracer
from opentelemetry.util.types import AttributeValue
from postgrest._async.request_builder import (
    AsyncQueryRequestBuilder,
    AsyncSingleRequestBuilder,
)
from postgrest._sync.request_builder import (
    SyncQueryRequestBuilder,
    SyncSingleRequestBuilder,
)
from wrapt import wrap_function_wrapper as _wrap

from .package import _instruments
from .utils import (
    parse_prefer_header_to_attributes,
    parse_sql_command,
    parse_sql_statement,
    parse_table_name,
)

logger = logging.getLogger(__name__)

_DEFAULT_SPAN_NAME_PREFIX = "supabase"
_DEFAULT_COMMAND_NAME = "request"


class SupabaseInstrumentor(BaseInstrumentor):
    def __init__(self, span_name_prefix=None):
        if not span_name_prefix:
            span_name_prefix = os.getenv(
                "OTEL_PYTHON_SUPABASE_SPAN_NAME_PREFIX",
                _DEFAULT_SPAN_NAME_PREFIX,
            )
        self._span_name_prefix = span_name_prefix.strip()

        self._should_record_statement = os.getenv("OTEL_PYTHON_SUPABASE_RECORD_STATEMENT", "true").lower() == "true"

        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        """Instrument Supabase client.

        Args:
            tracer_provider: The `TracerProvider` to use. If none is passed the
                current configured one is used.
        """
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(
            instrumenting_module_name=__name__,
            instrumenting_library_version=None,
            tracer_provider=tracer_provider,
            schema_url="https://opentelemetry.io/schemas/1.11.0",
        )

        _wrap(
            postgrest._sync.request_builder,
            "SyncQueryRequestBuilder.execute",
            _wrap_execute(
                tracer,
                self._span_name_prefix,
                self._should_record_statement,
            ),
        )
        _wrap(
            postgrest._sync.request_builder,
            "SyncSingleRequestBuilder.execute",
            _wrap_execute(
                tracer,
                self._span_name_prefix,
                self._should_record_statement,
            ),
        )
        _wrap(
            postgrest._async.request_builder,
            "AsyncQueryRequestBuilder.execute",
            _wrap_execute(
                tracer,
                self._span_name_prefix,
                self._should_record_statement,
            ),
        )
        _wrap(
            postgrest._async.request_builder,
            "AsyncSingleRequestBuilder.execute",
            _wrap_execute(
                tracer,
                self._span_name_prefix,
                self._should_record_statement,
            ),
        )

    def _uninstrument(self, **kwargs):
        unwrap(SyncQueryRequestBuilder, "execute")
        unwrap(SyncSingleRequestBuilder, "execute")
        unwrap(AsyncQueryRequestBuilder, "execute")
        unwrap(AsyncSingleRequestBuilder, "execute")


def _wrap_execute(
    tracer: Tracer,
    span_name_prefix: str,
    should_record_statement: bool = True,
):
    def wrapper(
        # TODO: proper callable type from postgrest request builder
        wrapped,
        builder: Union[
            SyncQueryRequestBuilder,
            SyncSingleRequestBuilder,
            AsyncQueryRequestBuilder,
            AsyncSingleRequestBuilder,
        ],
        args,
        kwargs,
    ):
        table_name = parse_table_name(builder.path)
        url = builder.session.base_url
        http_method = builder.http_method
        params = builder.params
        headers_attributes = parse_prefer_header_to_attributes(builder.headers)
        sql_command = parse_sql_command(http_method, headers_attributes)
        body = builder.json

        span_name = " ".join(
            s
            for s in [
                span_name_prefix,
                table_name,
                sql_command or http_method or _DEFAULT_COMMAND_NAME,
            ]
            if s
        )
        span_kind = SpanKind.CLIENT

        with tracer.start_as_current_span(
            name=span_name,
            kind=span_kind,
        ) as span:
            # extract attributes from req
            if span.is_recording():
                attributes: Dict[str, AttributeValue] = {
                    db_attributes.DB_SYSTEM: "postgresql",
                }

                if table_name:
                    attributes[db_attributes.DB_SQL_TABLE] = table_name

                if sql_command:
                    attributes[db_attributes.DB_OPERATION] = sql_command
                elif http_method:
                    attributes[http_attributes.HTTP_REQUEST_METHOD] = http_method

                if url:
                    attributes[server_attributes.SERVER_ADDRESS] = str(url)

                # insert/update payload
                if body:
                    # TODO: consider setting sanitized body payload as attribute
                    logger.debug("Sending Supabase payload: %s", body)

                # where, order by, etc.
                # TODO: should be db.statement. but might be impossible to reconstruct
                if should_record_statement and params:
                    attributes[db_attributes.DB_STATEMENT] = parse_sql_statement(params)

                if headers_attributes:
                    attributes.update(headers_attributes)

                for key, value in attributes.items():
                    span.set_attribute(key, value)

            # extract attributes from res
            res = wrapped(*args, **kwargs)

            data = getattr(res, "data", None)
            error = getattr(res, "error", None)
            # NOTE: res.count is often None, so disregard
            if span.is_recording():
                if data:
                    span.set_attribute("db.num_rows", len(data))
                if error:
                    span.set_status(
                        Status(status_code=StatusCode.ERROR, description=error)
                    )

            return res

    return wrapper



import traceback
from logging import LogRecord
from typing import Dict

from opentelemetry.semconv.trace import SpanAttributes

# skip natural LogRecord attributes
# http://docs.python.org/library/logging.html#logrecord-attributes
_RESERVED_ATTRS = frozenset(
    (
        "asctime",
        "args",
        "created",
        "exc_info",
        "exc_text",
        "filename",
        "funcName",
        "message",
        "levelname",
        "levelno",
        "lineno",
        "module",
        "msecs",
        "msg",
        "name",
        "pathname",
        "process",
        "processName",
        "relativeCreated",
        "stack_info",
        "thread",
        "threadName",
        "taskName",
    )
)


def get_attributes(record: LogRecord):
    """Monkeypatch for non-primitive (e.g. dict) attribute values defined in LogRecord.extra.

    See https://github.com/open-telemetry/opentelemetry-python/blob/c06e6f4b8616618907d70fa023eb2baab7a6ca61/opentelemetry-sdk/src/opentelemetry/sdk/_logs/_internal/__init__.py#L460
    """
    attributes = flatten_attributes(record)

    # Add standard code attributes for logs.
    attributes[SpanAttributes.CODE_FILEPATH] = record.pathname
    attributes[SpanAttributes.CODE_FUNCTION] = record.funcName
    attributes[SpanAttributes.CODE_LINENO] = record.lineno

    if record.exc_info:
        exctype, value, tb = record.exc_info
        if exctype is not None:
            attributes[SpanAttributes.EXCEPTION_TYPE] = exctype.__name__
        if value is not None and value.args:
            attributes[SpanAttributes.EXCEPTION_MESSAGE] = str(value.args[0])
        if tb is not None:
            # https://github.com/open-telemetry/opentelemetry-specification/blob/9fa7c656b26647b27e485a6af7e38dc716eba98a/specification/trace/semantic_conventions/exceptions.md#stacktrace-representation
            attributes[SpanAttributes.EXCEPTION_STACKTRACE] = "".join(
                traceback.format_exception(*record.exc_info)
            )
    return attributes


def flatten_attributes(record: LogRecord):
    """Convert LogRecord to flattened attribute dict.

    For instance {"a": {"b": 1}} will be converted to {"a.b": 1}.
    """
    seen = set()
    attrs = {}
    def flatten_helper(item, prefix=""):
        if prefix in _RESERVED_ATTRS:
            return
        if not isinstance(item, dict):
            attrs[prefix] = str(item)
            return
        for key, value in item.items():
            full_key = f"{prefix}.{key}" if prefix else key
            if id(value) in seen:
                # circular reference
                attrs[full_key] = str(value)
                continue
            seen.add(id(value))
            flatten_helper(value, full_key)
            seen.remove(id(value))
    flatten_helper(vars(record))
    return attrs



from .config import IudexConfig, configure_logger
from .instrumentation import instrument
from .fastapi import instrument_fastapi

__all__ = ["IudexConfig", "configure_logger", "instrument", "instrument_fastapi"]



import importlib.util
import logging
from typing import Optional, Union

from .config import IudexConfig, _IudexConfig

logger = logging.getLogger(__name__)


def instrument(
    service_name: Optional[str] = None,
    instance_id: Optional[str] = None,
    iudex_api_key: Optional[str] = None,
    log_level: Optional[Union[int, str]] = None,
    git_commit: Optional[str] = None,
    github_url: Optional[str] = None,
    env: Optional[str] = None,
    config: Optional[IudexConfig] = None,
):
    """Auto-instruments app to send OTel signals to Iudex.

    Invoke this function in your app entrypoint.

    Args:
        service_name: Name of the service, e.g. "billing_service", __name__.
            If not supplied, env var OTEL_SERVICE_NAME will be used.
        instance_id: ID of the service instance, e.g. container id, pod name.
        iudex_api_key: Your Iudex API key.
            If not supplied, env var IUDEX_API_KEY will be used.
        log_level: Logging level for the root logger.
        git_commit: Commit hash of the currently deployed code.
            Used with github_url to deep link telemetry to source code.
        github_url: URL of the GitHub repository.
            Used with git_commit to deep link telemetry to source code.
        env: Environment of the service, e.g. "production", "staging".
        config: IudexConfig object with more granular options.
            Will override all other args, so provide them to the object instead.
    """
    config = config or {
        "iudex_api_key": iudex_api_key,
        "service_name": service_name,
        "instance_id": instance_id,
        "logs_endpoint": None,
        "traces_endpoint": None,
        "log_level": log_level,
        "git_commit": git_commit,
        "github_url": github_url,
        "env": env,
    }
    iudex_config = _IudexConfig(**config)
    iudex_config.configure()

    _instrument_openai()
    _instrument_supabase()

    return iudex_config


def _instrument_openai():
    try:
        if importlib.util.find_spec("openai") is None:
            return
        from opentelemetry.instrumentation.openai import OpenAIInstrumentor

        instrumentor = OpenAIInstrumentor(
            enrich_assistant=True,
            enrich_token_usage=True,
            exception_logger=logger.error,
        )
        if not instrumentor.is_instrumented_by_opentelemetry:
            instrumentor.instrument()
    except Exception as e:
        logger.error(f"Failed to instrument OpenAI: {e}")

def _instrument_supabase():
    try:
        if importlib.util.find_spec("supabase") is None:
            return
        if importlib.util.find_spec("postgrest") is None:
            return
        from .supabase import SupabaseInstrumentor
        instrumentor = SupabaseInstrumentor()
        if not instrumentor.is_instrumented_by_opentelemetry:
            instrumentor.instrument()
    except Exception as e:
        logger.exception(e)
